Copyright (c) 2006 Carlos Ungil

Permission is hereby granted, free of charge, to any person obtaining
a copy of this software and associated documentation files (the
"Software"), to deal in the Software without restriction, including
without limitation the rights to use, copy, modify, merge, publish,
distribute, sublicense, and/or sell copies of the Software, and to
permit persons to whom the Software is furnished to do so, subject to
the following conditions:

The above copyright notice and this permission notice shall be
included in all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE
LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION
OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION
WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

(in-package :rcl)

(defvar *extract-single-element* t)

(defun r-type-decode (n)
  (cdr (assoc n *r-types*)))

(defun r-type (sexp)
  (let ((sxpinfo (sxpinfo-decode (sxpinfo-bitfield (sexp-sxpinfo sexp)))))
    (r-type-decode (first sxpinfo))))

  ;; the first version worked on PPC, but not on Intel
  ;; ppc:  671088768   #b 00101 0 00 0000000000000000 1 0 0 0 0 000
  ;; intel: 16777221   #b 000 0 0 0 0 1 0000000000000000 00 0 00101
  ;; there is something wrong with openmcl
  ;;       134217728   #b 00001000 00000000 00000000 00000000 
(defun sxpinfo-decode (int)
  (let (type obj named gp mark debug trace fin gcgen gccls)
    #+cffi-features:ppc32
    (progn
      ;; (5 1 2 16 1 1 1 1 1 3)
      (setf gccls (logand 7 int))
      (setf int (ash int -3))
      (setf gcgen (logand 1 int))
      (setf int (ash int -1))
      (setf fin (logand 1 int))
      (setf int (ash int -1))
      (setf trace (logand 1 int))
      (setf int (ash int -1))
      (setf debug (logand 1 int))
      (setf int (ash int -1))
      (setf mark (logand 1 int))
      (setf int (ash int -1))
      (setf gp (logand (1- (expt 2 16)) int))
      (setf int (ash int -16))
      (setf named (logand 3 int))
      (setf int (ash int -2))
      (setf obj (logand 1 int))
      (setf int (ash int -1))
      (setf type (logand (1- (expt 2 5)) int)))
    #-cffi-features:ppc32
    (progn
      ;; (5 1 2 16 1 1 1 1 1 3)
      (setf type (logand (1- (expt 2 5)) int))
      (setf int (ash int -5))
      (setf obj (logand 1 int))
      (setf int (ash int -1))
      (setf named (logand 3 int))
      (setf int (ash int -2))
      (setf gp (logand (1- (expt 2 16)) int))
      (setf int (ash int -16))
      (setf mark (logand 1 int))
      (setf int (ash int -1))
      (setf debug (logand 1 int))
      (setf int (ash int -1))      
      (setf trace (logand 1 int))
      (setf int (ash int -1))
      (setf fin (logand 1 int))
      (setf int (ash int -1))
      (setf gcgen (logand 1 int))
      (setf int (ash int -1))
      (setf gccls (logand 7 int)))
    (list type obj named gp mark debug trace fin gcgen gccls)))

(defun r-obj-describe (sexp)
  (let ((sxpinfo (sxpinfo-decode (sxpinfo-bitfield (sexp-sxpinfo sexp)))))
    (let ((type (r-type-decode (first sxpinfo))))
      (if (member type *r-vector-types*)
	  (let ((vecsxp (sexp-vecsxp sexp)))
	    (list type (vecsxp-length vecsxp) (vecsxp-true-length vecsxp)))
	  type))))

(defun r-obj-decode (sexp)
  ;; I don't know how to handle attributes, 
  ;; for the time being just print them
  (unless (eq :null (r-type (sexp-attrib sexp)))
    (print (r-obj-decode (sexp-attrib sexp))))
  (let ((result
	 (case (r-type sexp)
	   (:list-of-dotted-pairs
	    (let ((list (sexp-union sexp)))
	      (mapcar #'r-obj-decode
		      (list (listsxp-car list) (listsxp-cdr list) (listsxp-tag list)))))
	   (:null nil)
	   (:scalar-string-type (cffi:foreign-string-to-lisp (cffi:inc-pointer sexp 24)))
	   (:generic-vector (mapcar #'r-obj-decode (get-data-sexps sexp)))
	   (:logical-vector (mapcar #'plusp (get-data-integers sexp)))
	   (:string-vector (get-data-strings sexp))
	   (:real-vector (get-data-reals sexp))
	   (:integer-vector (get-data-integers sexp))
	   (t (r-obj-describe sexp)))))
    (if (and *extract-single-element* (listp result) (car result) (not (cdr result)))
	(car result)
	result)))
    

